/*
 * document.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "wwwio.h"

struct name
{
  char *name;
  void (*func)();
  void (*ifunc)();
};

static struct name protocols[] = 
{
  { "ftp", ftp, ftp_info },
  { "http", http, http_info },
  { "gopher", gopher, gopher_info },
  { "gopherp", gopherp, gopherp_info },
  { "file", file, file_info },
  { NULL, NULL, NULL },
};

/*
 * ProxyProtocol
 *
 * Deal with proxy stuff
 */
static int
ProxyProtocol(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  int plen, hlen;
  Document *d = NULL;
  URLParts *rup;
  char *pproxy = NULL;
  char *no_proxy;
  char *cp, *ph;
  char *pname;

  /*
   * Hee, allow any weird protocol to be checked for proxy.
   */
  pname = (char *)alloc_mem(strlen(up->protocol) + strlen("_proxy") + 1);
  sprintf (pname, "%s_proxy", up->protocol);

  if ((pproxy = getenv(pname)) == NULL)
  {
    pproxy = NGetFromStringDB(pname);
    if (pproxy == NULL)
    {
      if ((pproxy = getenv("all_proxy")) == NULL)
      {
	pproxy = NGetFromStringDB("all_proxy");
	if (pproxy == NULL)
	{
	  free_mem(pname);
	  return(0);
	}
      }
    }
  }
  pproxy = alloc_string(pproxy);
  free_mem(pname);

  /* 
   * Check for domains that are not allowed for proxy
   */
  if ((no_proxy = getenv("no_proxy")) != NULL
      || (no_proxy = NGetFromStringDB("no_proxy")) != NULL)
  {
    hlen = strlen(up->hostname);
    cp = no_proxy;
    while ((ph = mystrtok(cp, ',', &cp)) != NULL)
    {
      plen = strlen(ph);
      if (plen <= hlen &&
	  strncasecmp(ph, up->hostname + (hlen - plen), plen) == 0)
      {
	free_mem(pproxy);
	return(0);
      }
    }
  }
  
  /*
   * Try to get the proxy connection going.
   */
  if ((rup = ParseURL(pproxy)) != NULL)
  {
    char *msg = GetFromStringDB("absrurl");
    
    if (IsAbsoluteURL(rup))
    {
      http_proxy(ioc, rup, up, func, closure);
    }
    else
    {
      d = BuildDocument(msg, strlen(msg), "text/html", 1, 0);
      d->status = DS_ERROR;
      (func)(d, closure, 1);
    }
    
    DestroyURLParts(rup);
  }
  else
  {
    char *msg = GetFromStringDB("invurl");
    d = BuildDocument(msg, strlen(msg), "text/html", 1, 0);
    d->status = DS_ERROR;
    (func)(d, closure, 1);
  }

  free_mem(pproxy);

  return(1);
}

/*
 * LocalProtocol
 *
 * Handle protocols that have been implemented in the chimera source.
 */
static void
LocalProtocol(ioc, up, func, closure, infoonly)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
int infoonly;
{
  int i;

  for (i = 0; protocols[i].name != NULL; i++)
  {
    if (strcasecmp(protocols[i].name, up->protocol) == 0)
    {
      if (infoonly != 0) (protocols[i].ifunc)(ioc, up, func, closure);
      else (protocols[i].func)(ioc, up, func, closure);
      break;
    }
  }

  return;
}

/*
 * CreateDocument
 *
 * Allocates the structure used for documents.  Could do other
 * init stuff, too.  This function should always be called to create
 * a new document.  It provides initialization for the document which
 * is good if an old part of the program doesn't know about new
 * document attributes.
 */
Document *
CreateDocument()
{
  Document *d;

  d = (Document *)alloc_mem(sizeof(Document));

  d->text = NULL;
  d->len = 0;

  d->from_cache = 0;
  d->cache = 1;

  d->status = DS_OK;

  d->mflist = NULL;

  d->decode = NULL;

  return(d);
}

/*
 * DestroyDocument
 */
void
DestroyDocument(d)
Document *d;
{
  MIMEField *m, *t;

  return;

  if (d->text != NULL) free_mem(d->text);

  for (m = d->mflist; m != NULL; )
  {
    t = m;
    m = m->next;
    MIMEDestroyField(t);
  }

  free_mem((char *)d);

  return;
}

/*
 * BuildDocument
 *
 * Do some of the grunge work to make a document.
 */
Document *
BuildDocument(text, len, content, doalloc, cache)
char *text;
int len;
char *content;
int doalloc;
int cache;
{
  Document *d;

  d = CreateDocument();
  if (doalloc && text != NULL)
  {
    d->text = alloc_mem(len + 1);
    memcpy(d->text, text, len);
    d->text[len] = '\0';
  }
  else
  {
    d->text = text;
  }
  d->len = (text == NULL ? 0:len);

  if (content != NULL) MIMEAddHeaderField(d, "Content-type", content);

  d->cache = cache;
  d->from_cache = 0;

  return(d);
}

/*
 * LoadDocument
 */
void
LoadDocument(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  if (!IsAbsoluteURL(up))
  {
    char *msg = GetFromStringDB("absurl");
    Document *d;

    d = BuildDocument(msg, strlen(msg), "text/html", 1, 0);
    d->status = DS_ERROR;
    (func)(d, closure, 1);
    return;
  }

  if (ProxyProtocol(ioc, up, func, closure) == 0)
  {
    LocalProtocol(ioc, up, func, closure, 0);
  }

  return;
}

/*
 * LoadDocumentInfo
 */
void
LoadDocumentInfo(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  LocalProtocol(ioc, up, 1, func, closure, 1);
  return;
}
