/*
 * http.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "wwwio.h"
#include "wwwioextern.h"
#include "net.h"

#define DEFAULT_HTTP_PORT 80

typedef struct _httpinfo
{
  Document *d;
  int (*func)();
  char *closure;
  URLParts *up;
  URLParts *rup;
  char *request;
  int infoonly;
  IOContext ioc;
} HTTPInfo;

static char *build_request _ArgProto((URLParts *, char *, char *));

/*
 * ECToDS
 */
static void
ECToDS(d, error_code)
Document *d;
int error_code;
{
  if (error_code <= 200 && error_code < 300)
  {
    if (error_code == 204) d->status = DS_NOTHING;
    else d->status = DS_OK;
  }
  else if (error_code >= 300 && error_code < 400)
  {
    if (error_code == 302) d->cache = 0;
  }
  else if (error_code >= 400 && error_code < 500)
  {
    if (error_code == 401) d->status = DS_NEEDS_AUTH;
    else d->status = DS_ERROR;
  }
  else
  {
    d->status = DS_ERROR;
  }
  return;
}

static char *format_get = "\
GET %s HTTP/1.0\r\n\
User-Agent: %s\r\n\
Accept: */*\r\n\
%s\
%s\
\r\n";

static char *format_get2 = "\
GET %s?%s HTTP/1.0\r\n\
User-Agent: %s\r\n\
Accept: */*\r\n\
%s\
%s\
\r\n";

static char *format_post = "\
POST %s HTTP/1.0\r\n\
User-Agent: %s\n\
Accept: */*\r\n\
Content-length: %d\r\n\
Content-type: %s\r\n\
%s\
%s\
\r\n";

/*
 * crap for uuencode
 */
static char six2pr[64] =
{
    'A','B','C','D','E','F','G','H','I','J','K','L','M',
    'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
    'a','b','c','d','e','f','g','h','i','j','k','l','m',
    'n','o','p','q','r','s','t','u','v','w','x','y','z',
    '0','1','2','3','4','5','6','7','8','9','+','/'
};

/*
 * uuencode
 *
 * This is used to uuencode a string.  It pisses me off that this is even
 * needed.  The WWW guys LIKE complexity and huge code because it keeps
 * them employed (ah wild speculation is good!).
 *
 * I snarfed this code from some version of libwww.
 *
 * ACKNOWLEDGEMENT:
 *      This code is taken from rpem distribution, and was originally
 *      written by Mark Riordan.
 *
 * AUTHORS:
 *      MR      Mark Riordan    riordanmr@clvax1.cl.msu.edu
 *      AL      Ari Luotonen    luotonen@dxcern.cern.ch
 *
 */
static int
uuencode(bufin, nbytes, bufcoded)
unsigned char *bufin;
unsigned int nbytes;
char *bufcoded;
{
  /* ENC is the basic 1 character encoding function to make a char printing */
#define ENC(c) six2pr[c]
  
  register char *outptr = bufcoded;
  unsigned int i;
  
  for (i=0; i<nbytes; i += 3) 
  {
    *(outptr++) = ENC(*bufin >> 2);            /* c1 */
    *(outptr++) = ENC(((*bufin << 4) & 060) | ((bufin[1] >> 4) & 017)); /*c2*/
    *(outptr++) = ENC(((bufin[1] << 2) & 074) | ((bufin[2] >> 6) & 03));/*c3*/
    *(outptr++) = ENC(bufin[2] & 077);         /* c4 */
    
    bufin += 3;
  }
  
  /* If nbytes was not a multiple of 3, then we have encoded too
   * many characters.  Adjust appropriately.
   */
  if (i == nbytes + 1)
  {
    /* There were only 2 bytes in that last group */
    outptr[-1] = '=';
  }
  else if (i == nbytes + 2)
  {
    /* There was only 1 byte in that last group */
    outptr[-1] = '=';
    outptr[-2] = '=';
  }
  *outptr = '\0';

  return(outptr - bufcoded);
}

/*
 * build_auth_info
 *
 * Build the string needed for authorization.
 */
static char *
build_auth_info(type, username, password)
char *type;
char *username;
char *password;
{
  char *garbage;
  char *auth_info;
  static char *format = "Authorization: %s %s\n";

  garbage = alloc_mem(strlen(username) +
		      1 +
		      strlen(password != NULL ? password:"") +
		      1);
  strcpy(garbage, username);
  if (password != NULL) strcat(garbage, ":");
  if (password != NULL) strcat(garbage, password);

  if (strcmp(type, "Basic") == 0)
  {
    char *t;

    t = alloc_mem(strlen(garbage) * 2);
    uuencode(garbage, strlen(garbage), t);
    free_mem((char *)garbage);
    garbage = t;;
  }

  auth_info = alloc_mem(strlen(format) +
			strlen(type) +
			strlen(garbage) + 1);
  sprintf (auth_info, format,
	   type,
	   garbage);

  return(auth_info);
}

/* 
 * build_request
 *
 * Make a HTTP request string
 */
static char *
build_request(up, filename, extra_header)
URLParts *up;
char *filename;
char *extra_header;
{
  char *query;
  char *auth_info = NULL;

  if (filename == NULL) filename = up->filename;
  if (extra_header == NULL) extra_header = "";

  if (up->auth_type != NULL)
  {
    if (up->username != NULL)
    {
      auth_info = build_auth_info(up->auth_type, up->username, up->password);
    }
  }
  if (auth_info == NULL) auth_info = alloc_string("");

  if (up->input_data != NULL)
  {
    if (up->method == NULL || strcasecmp(up->method, "GET") == 0) /* GET ! */
    {
      query = alloc_mem(strlen(filename) +
			strlen(format_get2) +
			strlen(up->input_data) +
			strlen(extra_header) +
			strlen(auth_info) + 
			strlen(USER_AGENT) + 1);
      sprintf (query,
	       format_get2,
	       filename,
	       up->input_data,
	       USER_AGENT,
	       extra_header,
	       auth_info);
    }
    else if (strcasecmp(up->method, "POST") == 0) /* POST ! */
    {
      query = alloc_mem(strlen(filename) +
			strlen(format_post) +
			up->input_data_len +
			strlen(up->input_data_type) +
			strlen(extra_header) +
			strlen(auth_info) +
			strlen(USER_AGENT) + 1);
      sprintf (query,
	       format_post,
	       filename,
	       USER_AGENT,
	       up->input_data_len,
	       up->input_data_type,
	       extra_header,
	       auth_info);
      memcpy(query + strlen(query), up->input_data, up->input_data_len);
    }
    else query = NULL;
  }
  else /* regular GET */
  {
    query = alloc_mem(strlen(filename) +
		      strlen(format_get) +
		      strlen(extra_header) +
		      strlen(auth_info) + 
		      strlen(USER_AGENT) + 1);
    sprintf (query,
	     format_get,
	     filename,
	     USER_AGENT,
	     extra_header,
	     auth_info);
  }

  free_mem((char *)auth_info);

  return(query);
}

/*
 * make_http_request
 *
 * Builds an HTTP request
 */
static char *
make_http_request(up, type, infoonly)
URLParts *up;
int type;
int infoonly;
{
  char *query;
  char *extra_header;

  if (type == 0)
  {
    if (up->reload != 0) extra_header = "Pragma: no-cache\r\n";
    else extra_header = "";

    query = build_request(up, NULL, extra_header);
  }
  else if (type == 1)
  {
    char *filename;

    if (up->reload != 0) extra_header = "Pragma: no-cache\r\n";
    else extra_header = "";

    filename = MakeURL(up, 0);
    query = build_request(up, filename, extra_header);
    free_mem((char *)filename);
  }
  else
  {
    char *url;
    static char *format = "\
URI: %s\r\n\
X-protocol: %s\r\n\
X-hostname: %s\r\n\
X-port: %d\r\n\
X-filename: %s\r\n";

    url = MakeURL(up, 0);
    extra_header = alloc_mem(strlen(url) * 5 + strlen(format) + 1);
    sprintf (extra_header,
	     format,
	     url,
	     up->protocol != NULL ? up->protocol:"",
	     up->hostname != NULL ? up->hostname:"",
	     up->port,
	     up->filename != NULL ? up->filename:"/");

    query = build_request(up, NULL, extra_header);
    free_mem((char *)extra_header);
    free_mem((char *)url);
  }

  return(query);
}

/*
 * http_destroy_info
 */
static void
http_destroy_info(hi)
HTTPInfo *hi;
{
  if (hi->request != NULL) free_mem(hi->request);
  if (hi->up != NULL) DestroyURLParts(hi->up);
  if (hi->rup != NULL) DestroyURLParts(hi->rup);

  free_mem((char *)hi);

  return;
}

/*
 * http_clean
 */
static void
http_clean(s, closure, status)
int s;
char *closure;
int status;
{
  HTTPInfo *hi = (HTTPInfo *)closure;

  if (s >= 0) IOEnd(hi->ioc, s);

  hi->d->status = status;
  if (hi->func != NULL) (hi->func)(hi->d, hi->closure, -1);
  http_destroy_info(hi);

  return;
}

/*
 * http_data
 */
static void
http_data(s, t, tlen, eof, closure)
int s;
char *t;
int tlen;
int eof;
char *closure;
{
  HTTPInfo *hi = (HTTPInfo *)closure;

  hi->d->text = t;
  hi->d->len = tlen;

  if (eof)
  {
    IOEnd(hi->ioc, s);
    (hi->func)(hi->d, hi->closure, 1);
    http_destroy_info(hi);
  }
  else
  {
    if ((hi->func)(hi->d, hi->closure, 0) < 0)
    {
      IOEnd(hi->ioc, s);
      http_destroy_info(hi);
    }
    else
    {
      IOReadStream(hi->ioc, s, http_data);
    }
  }

  return;
}

/*
 * http_header
 */
void
http_header(s, t, closure)
int s;
char *t;
char *closure;
{
  HTTPInfo *hi = (HTTPInfo *)closure;
  char *value;

  /*
   * End of header?
   */
  if (t[0] == '\r' || t[0] == '\n')
  {
    if (hi->infoonly != 0)
    {
      IOEnd(hi->ioc, s);
      (hi->func)(hi->d, hi->closure, 0);
      http_destroy_info(hi);

      return;
    }
    else
    {
      if ((hi->func)(hi->d, hi->closure, 0) < 0)
      {
	IOEnd(hi->ioc, s);
	http_destroy_info(hi);
      }
      else
      {
	int len;

	if ((value = MIMEGetFieldValue(hi->d->mflist,
				       "content-length", 0)) != NULL)
	{
	  if ((len = atoi(value)) < 0) len = 0;
	}
	else len = 0;

	IOResetStream(hi->ioc, s, len);
	IOReadStream(hi->ioc, s, http_data);
      }
    }
  }
  else
  {
    MIMEParseHeaderLine(hi->d, t);
    IOReadLine(hi->ioc, s, http_header);
  }

  return;
}

/*
 * http_status
 */
static void
http_status(s, t, closure)
int s;
char *t;
char *closure;
{
  HTTPInfo *hi = (HTTPInfo *)closure;

  ECToDS(hi->d, atoi(t));

  IOReadLine(hi->ioc, s, http_header);

  return;
}

/*
 * http_check
 */
static void
http_check(s, t, tlen, eof, closure)
int s;
char *t;
int tlen;
int eof;
char *closure;
{
  HTTPInfo *hi = (HTTPInfo *)closure;

  /*
   * Check to see if this is an HTTP/1.0 server.  If not then
   * fake a content type else go on to read an HTTP/1.0 header.
   */
  if (tlen < 8 || strncmp(t, "HTTP/1.0", 8) != 0)
  {
    char *content;

    if ((content = MIMEExt2Content(hi->up->filename)) == NULL)
    {
      content = "text/html";
    }

    MIMEAddHeaderField(hi->d, "content-type", content);

    if ((hi->func)(hi->d, hi->closure, tlen < 8 ? 1:0) < 0)
    {
      IOEnd(hi->ioc, s);
      http_destroy_info(hi);
    }
    else
    {
      IOSetMax(hi->ioc, s, 0);
      IOReadStream(hi->ioc, s, http_data);
    }
  }
  else
  {
    IOReadLine(hi->ioc, s, http_status);
  }

  return;
}

/*
 * http_conect
 */
static void
http_connect(s, closure)
int s;
char *closure;
{
  HTTPInfo *hi = (HTTPInfo *)closure;

  IOSetMax(hi->ioc, s, 8);
  IOReadStream(hi->ioc, s, http_check);

  return;
}

/*
 * http_write
 */
static void
http_write(s, closure)
int s;
char *closure;
{
  HTTPInfo *hi = (HTTPInfo *)closure;

  IOConnectStream(hi->ioc, s, http_connect);

  return;
}

/*
 * http_start
 */
static void
http_start(closure, status)
void *closure;
int status;
{
  HTTPInfo *hi = (HTTPInfo *)closure;
  int s;
  URLParts *up = hi->up;
  URLParts *rup = hi->rup;

  if (status == -1)
  {
    http_destroy_info(hi);
  }
  else
  {
    s = net_open(up->hostname != NULL ? up->hostname:"www",
		 up->port == 0 ? DEFAULT_HTTP_PORT:up->port);
    if (s < 0) http_destroy_info(hi);
    else
    {
      IORegister(hi->ioc, s, http_clean, (char *)hi);
      
      if (rup != NULL)
      {
	hi->request = make_http_request(rup, 1, hi->infoonly);
      }
      else
      {
	hi->request = make_http_request(up, 0, hi->infoonly);
      }
      
      IOWriteStream(hi->ioc, s, hi->request, strlen(hi->request), http_write);
    }
  }

  return;
}

/*
 * http_main
 *
 * Talk to an HTTP server.
 */
static void
http_main(ioc, up, rup, func, closure, infoonly)
IOContext ioc;
URLParts *up;
URLParts *rup;
int (*func)();
char *closure;
int infoonly;
{
  HTTPInfo *hi;

  hi = (HTTPInfo *)alloc_mem(sizeof(HTTPInfo));
  hi->d = CreateDocument();
  hi->func = func;
  hi->closure = closure;
  hi->up = DupURLParts(up);
  if (rup != NULL) hi->rup = DupURLParts(rup);
  else hi->rup = NULL;
  hi->request = NULL;
  hi->infoonly = 0;
  hi->ioc = ioc;

  IOStart(ioc, http_start, hi);

  return;
}
  
/*
 * http
 *
 * Frontend for regular HTTP
 */
void
http(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  http_main(ioc, up, NULL, func, closure, 0);

  return;
}

/*
 * http_info
 *
 * Just do a HEAD
 */
void
http_info(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  http_main(ioc, up, NULL, func, closure, 1);

  return;
}

/*
 * http_proxy
 *
 * Frontend for proxy HTTP
 */
void
http_proxy(ioc, up, rup, func, closure)
IOContext ioc;
URLParts *up;
URLParts *rup;
int (*func)();
void *closure;
{
  http_main(ioc, up, rup, func, closure, 0);

  return;
}
