/* dither.c
 *
 * Thrashed by john@isri.unlv.edu.
 *
 * completely reworked dithering module for xloadimage
 * uses error-diffusion dithering (floyd-steinberg) instead
 * of simple 4x4 ordered-dither that was previously used
 *
 * the previous version of this code was written by steve losen
 * (scl@virginia.edu)
 * 
 * jim frost    07.10.89
 * Steve Losen  11.17.89
 * kirk johnson 06.04.90
 *
 * Copyright 1990 Kirk L. Johnson (see the included file
 * "kljcpyrght.h" for complete copyright information)
 *
 * Copyright 1989, 1990 Jim Frost and Steve Losen.
 * See included file "copyright.h" for complete copyright information.
 */
#include <stdio.h>

#include "common.h"
#include "copyright.h"
#include "kljcpyrght.h"
#include "xloadimage.h"
#include "dither.h"

#define MaxIntensity  65536	/* maximum possible Intensity */

#define MaxGrey       32768	/* limits on the grey levels used */
#define Threshold     16384	/* in the dithering process */
#define MinGrey           0

static unsigned int tone_scale_adjust();
static void         LeftToRight();
static void         RightToLeft();

/*
 * simple floyd-steinberg dither with serpentine raster processing
 */

/*
 * a _very_ simple tone scale adjustment routine. provides a piecewise
 * linear mapping according to the following:
 *
 *      input:          output:
 *     0 (MinGrey)    0 (MinGrey)
 *     Threshold      Threshold/2
 *     MaxGrey        MaxGrey
 * 
 * this should help things look a bit better on most displays.
 */
static unsigned int
tone_scale_adjust(val)
unsigned int val;
{
  unsigned int rslt;
  
  if (val < Threshold) rslt = val / 2;
  else
  {
    rslt = (((val - Threshold) * (MaxGrey-(Threshold/2))) /
	    (MaxGrey-Threshold)) + (Threshold/2);
  }
    
  return(rslt);
}


/*
 * dither a line from left to right
 */
static void
LeftToRight(curr, next, width)
int *curr;
int *next;
int  width;
{
  int idx;
  int error;
  int output;

  for (idx = 0; idx < width; idx++)
  {
    output       = (curr[idx] > Threshold) ? MaxGrey : MinGrey;
    error        = curr[idx] - output;
    curr[idx]    = output;
    next[idx-1] += error * 3 / 16;
    next[idx]   += error * 5 / 16;
    next[idx+1] += error * 1 / 16;
    curr[idx+1] += error * 7 / 16;
  }

  return;
}


/*
 * dither a line from right to left
 */
static void
RightToLeft(curr, next, width)
int *curr;
int *next;
int  width;
{
  int idx;
  int error;
  int output;

  for (idx = (width - 1); idx >= 0; idx--)
  {
    output       = (curr[idx] > Threshold) ? MaxGrey : MinGrey;
    error        = curr[idx] - output;
    curr[idx]    = output;
    next[idx+1] += error * 3 / 16;
    next[idx]   += error * 5 / 16;
    next[idx-1] += error * 1 / 16;
    curr[idx-1] += error * 7 / 16;
  }

  return;
}

/*
 * ditherInit
 */
ditherState *
ditherInit(cimage)
Image *cimage;
{
  ditherState *ds;
  int j;

  ds = (ditherState *)alloc_mem(sizeof(ditherState));

  ds->image = newBitImage(cimage->width, cimage->height);
  ds->cimage = cimage;

  ds->curr  = (int *)alloc_mem(sizeof(int) * (cimage->width + 2));
  ds->next  = (int *)alloc_mem(sizeof(int) * (cimage->width + 2));
  ds->curr += 1;
  ds->next += 1;
  for (j = 0; j < cimage->width; j++)
  {
    ds->curr[j] = 0;
    ds->next[j] = 0;
  }

  return(ds);
}

int
ditherAddLine(ds, line)
ditherState *ds;
int line;
{
  int j;
  Image *cimage = ds->cimage;
  Pixel color;
  byte *src;
  byte *dst;

  /* copy the row into the current line */
  src = cimage->data + cimage->pixlen * line * cimage->width;
  for (j = 0; j < cimage->width; j++)
  {
    color = memToVal(src, cimage->pixlen);
    src  += cimage->pixlen;
    
    ds->curr[j] +=
	tone_scale_adjust(colorIntensity(cimage->rgb.red[color],
					 cimage->rgb.green[color],
					 cimage->rgb.blue[color]) >> 1);
  }
  
  /* dither the current line */
  if (line & 0x01) RightToLeft(ds->curr, ds->next, cimage->width);
  else LeftToRight(ds->curr, ds->next, cimage->width);
  
  /* copy the dithered line to the destination image */
  dst = ds->image->data + line * ds->image->width;
  for (j = 0; j < cimage->width; j++, dst++)
  {
    if (ds->curr[j] < Threshold) *dst = 1;
  }
  
  /* circulate the line buffers */
  ds->swap = ds->curr;
  ds->curr = ds->next;
  ds->next = ds->swap;
  for (j = 0; j < cimage->width; j++)
  {
    ds->next[j] = 0;
  }

  return(0);
}

/*
 * ditherDestroy
 */
void
ditherDestroy(ds)
ditherState *ds;
{
  freeImage(ds->image);
  free_mem((char *)(ds->curr - 1));
  free_mem((char *)(ds->next - 1));
  free_mem((char *)ds);

  return;
}
